/*
 *     terraVox: an experimental voxel landscape engine
 *
 *     Copyright (C) 1999/2000 Alex J. Champandard
 *
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *     The tutorial that accompanies this source code can be found at
 *         http://www.flipcode.com
 *
 *     You can also visit the terraVox web site at
 *         http://atlas.cs.york.ac.uk/~ajc116/terraVox
 *
 *     And if you wish to contact me for any reason, I can be reached at
 *         alexjc@altavista.com
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <fstream.h>
#include "vector.h"
#include "matrix.h"
#include "keyboard.h"
#include "timer.h"
#include "vga.h"
#include "pack.h"


typedef unsigned char byte;
typedef unsigned short word;

// rendering parameters
long RenderSteps = 128;             // total number of steps
long RenderZone = 1024;

// the step size is interpolated, and stored for all RenderSteps in DistInfo[]
long NearStepSize = 256;         // 24.8 fixed point: step size 1
long FarStepSize = 2048;         // step size 4

// 128 - no fog
// 0 - black fog
word NearFogValue = 128<<8;
word FarFogValue = 0;
long NearFogStep = 64;         // when to start interpolating the fog value


// use 8x8 blocks for better caching
#define BlockSizeY   8
#define BlockSizeX   8
#define HighBlockMaskY 0xfffffff8
#define HighBlockMaskX 0xfffffff8
#define LowBlockMaskY 0x00000007
#define LowBlockMaskX 0x00000007
#define BlockShiftX 3
#define BlockShiftY 3

#define HeightMapScale    32
#define TextureScale      4               // how many voxels does one texture cover

#define SkyHeight    384                  // plane at Y=384
#define SkyScale     1024                 // sky texture scale

// 3D rendering parameters
#define HScale       320
#define VScale       -2000
#define HCenter      160
#define VCenter      100

// rendering flags
bool r_light = true;
bool r_fog   = true;
bool r_mip   = true;
bool r_bili  = false;
bool r_antia = false;
bool r_int   = false;

#define r_gourd (r_fog || r_light)

// options display
long optCount = 840;
byte optNum   = 84;

// time FPS
long FrameCount = 0;
long lastTime = 0;
long lastCount = 0;
long curFPS = 0;
// movement update
long lastUpdt = 0;

// interlacing
long Interlace = 0;


// some very usefull classes :)
VGA *vga;
KEYBOARD *keyboard;
TIMER *timer;
PACKFILE *packfile;

// misc. display
byte *Info, *Digits;
// pointers to maps
byte *HeightMap, *LightMap, *RefMap;
// precalculated tables
byte *PropTab, *MixTab;
// textures
byte *SkyText;
byte *Texture[4];
// occlusion buffer
long *cBuffer;

// 3d variables
matrix Camera, Inverse;
vector CamPosVoxSpace;
double AngX, AngY, AngZ, TrX, TrY, TrZ;

// all necessary info at each step
struct TDistInfo
{
       long Persp;
       word FogVal;
       long StepSize;
};
TDistInfo *DistInfo;

// precalculate sky projection
vector SkyRays[200][2];

// palettes
struct RGB { byte R, G, B; };
RGB PalSky[256], PalLand[256];
// have pals been modified?
bool Modified;

// misc.
long LastMove, Noise;

/*
 * without c-buffer (slow):
 *      draw the sky fullscreen, draw the voxel over the top.
 * with c-buffer:
 *      draw voxel updating the c-buffer, only fill the rest of the
 *      screen with the sky.
 */
void Reset_cBuffer()
{
 if (r_int) for (int i=0; i<160; i++) {
     cBuffer[(i<<1)+Interlace] = 200<<12;
     cBuffer[(i<<1)+1-Interlace] = 0;
 }
 else for (int i=0; i<320; i++) cBuffer[i] = 200<<12;
};

/*
 * precalculate coefficiants for bilinear filtering
 */
void Prepare_Bilinear()
{
   int i, j, k;
   for (j=0; j<256; j++)
   {
      for (i=0; i<256; i++)
      {
           if (i<128) k = (j*i/127);
           else k = j+((255-j)*(i-128))/128;
           MixTab[(j<<8)+i] = (byte)k;
      }
   }
   for (j=0; j<256; j++)
   {
      for (i=0; i<256; i++)
      {
          PropTab[(j<<8)+i] = (byte)((i*j)/255);
      }
   }
};

/*
 * get 8.8 fixed point bilinear filtered value
 */
word GetValue( byte *where, long tX, long tZ )
{
    long itZ = (tZ>>8)&0xff00;
    long itX = (tX>>16)&0x00ff;
    long dtZ = (tZ>>8)&0xff;
    long dtX = (tX>>8)&0xff;
    return ((where[itZ+itX] * (0xff-dtX) * (0xff-dtZ) )>>8)
         + ((where[((itZ+256)&0xff00)+itX] * (0xff-dtX) * dtZ )>>8)
         + ((where[((itZ+256)&0xff00)+((itX+1)&0x00ff)] * dtX * dtZ )>>8)
         + ((where[itZ+((itX+1)&0x00ff)] * dtX * (0xff-dtZ) )>>8);
};

byte addrblock( byte *where, long x, long y )
{
 long addr = ( ((y&HighBlockMaskY)<<10)
             + ((x&HighBlockMaskX)<<BlockShiftY)
             + ((y&LowBlockMaskY)<<BlockShiftY) + (x&LowBlockMaskX));
     return where[addr];
}

word Get_Value( byte *where, long tX, long tZ )
{
   long itZ = (tZ>>16)&0x3ff;
   long itX = (tX>>16)&0x3ff;
   long dtZ = (tZ>>8)&0xff;
   long dtX = (tX>>8)&0xff;
/*   return ((where[itZ+itX] * (0xff-dtX) * (0xff-dtZ) )>>8)
        + ((where[((itZ+1024)&0xffc00)+itX] * (0xff-dtX) * dtZ )>>8)
        + ((where[((itZ+1024)&0xffc00)+((itX+1)&0x003ff)] * dtX * dtZ )>>8)
        + ((where[itZ+((itX+1)&0x003ff)] * dtX * (0xff-dtZ) )>>8); */
   return ((addrblock( where, itX,itZ) * (0xff-dtX) * (0xff-dtZ) )>>8)
        + ((addrblock( where, itX,((itZ+1)&0x3ff)) * (0xff-dtX) * dtZ )>>8)
        + ((addrblock( where, ((itX+1)&0x3ff),((itZ+1)&0x3ff) )  * dtX * dtZ )>>8)
        + ((addrblock( where, ((itX+1)&0x3ff),itZ) * dtX * (0xff-dtZ) )>>8);
};

word GetTexel( byte index, long base, long x1, word xmask,
                                      long z1, word zmask )
{
   if (r_bili)
   {
    long itZ = ((z1>>TextureScale)&zmask);
    long itX = ((x1>>(8+TextureScale))&xmask);
    long dtZ = (z1>>TextureScale)&0xff;
    long dtX = (x1>>TextureScale)&0xff;
    return ((Texture[index][base+itZ+itX] * (0xff-dtX) * (0xff-dtZ) )>>8)
         + ((Texture[index][base+((itZ+256)&zmask)+itX] * (0xff-dtX) * dtZ )>>8)
         + ((Texture[index][base+((itZ+256)&zmask)+((itX+1)&xmask)] * dtX * dtZ )>>8)
         + ((Texture[index][base+itZ+((itX+1)&xmask)] * dtX * (0xff-dtZ) )>>8);
   } else return (word)Texture[index][base+((z1>>TextureScale)&zmask)+((x1>>(8+TextureScale))&xmask)]<<8;
}


/*
 * draw a vertical line
 */
void vLine( long x, long y1, word c1, long x1, long z1,
            long y2, word c2, long x2, long z2, int aac )
{
   word t1 = Get_Value( RefMap, x1, z1 );
   word t2 = Get_Value( RefMap, x2, z2 );
   long dx, dz, dt;
   long iy1 = (y1>>12);
   long iy2 = (y2>>12);
   long dy = iy2-iy1;
   long offs = (iy1<<8) + (iy1<<6) + x;
   byte lastval;
   if (dy>0)
   {
      dx = (x2-x1)/dy;
      dz = (z2-z1)/dy;
      dt = (t2-t1)/dy;
   }
   else {
      dx = 0; dz = 0; dt = 0;
   }
   long sizex = 128;
   long sizez = 128;
   long base = 0;
   bool scalex = false;
   bool scalez = false;
   if ((r_mip) && (dy>0))
   {
      while (abs(dz)>(65536>>TextureScale))
      {
            dz = dz>>1;
            z1 = z1>>1;
            base += (sizez<<8);
            sizez = sizez>>1;
            scalez = true;
      };
      while (abs(dx)>(65536>>TextureScale))
      {
            dx = dx>>1;
            x1 = x1>>1;
            base += sizex;
            sizex = sizex>>1;
            scalex = true;
      };
   };
   long zmask = (sizez-1)<<8;
   long xmask = (sizex-1);
   byte zshift = 0+TextureScale;
   byte xshift = 8+TextureScale;
   long texel;
   if (r_gourd) // mixed textured gouraud span
   {
       if (dy>0)
       {
          word dc = (c2-c1)/dy;
          for (long y=iy1; y<iy2; y++ )
          {
              texel = base+((z1>>zshift)&zmask)+((x1>>xshift)&xmask);
              vga->page_draw[offs] = (byte)MixTab[
               ( ( GetTexel( (t1>>8)&0x3, base, x1, xmask, z1, zmask )
                   * (0xff-(t1&0xff)) >> 8 )
               + ( GetTexel( ((t1>>8)+1)&0x3, base, x1, xmask, z1, zmask )
                   * (t1&0xff) >> 8 ) & 0xff00 ) + (c1>>8)];
              offs += 320;
              c1 += dc;
              x1 += dx;
              z1 += dz;
              t1 += dt;
          };
       }
       if (r_antia) lastval = (byte)MixTab[
          ( ( GetTexel( (t1>>8)&0x3, base, x1, xmask, z1, zmask )
             * (0xff-(t1&0xff)) >> 8 )
          + ( GetTexel( ((t1>>8)+1)&0x3, base, x1, xmask, z1, zmask )
             * (t1&0xff) >> 8 ) & 0xff00 ) + (c1>>8)];
   }
   else { // mixed textured span
          if (dy>0)
          {
              for (long y=iy1; y<iy2; y++ )
              {
                 texel = base+((z1>>zshift)&zmask)+((x1>>xshift)&xmask);
                 vga->page_draw[offs] = (byte)
                 ( GetTexel( (t1>>8)&0x3, base, x1, xmask, z1, zmask )
                   * (0xff-(t1&0xff)) >> 16 )
               + ( GetTexel( ((t1>>8)+1)&0x3, base, x1, xmask, z1, zmask )
                   * (t1&0xff) >> 16 );
                 offs += 320;
                 x1 += dx;
                 z1 += dz;
                 t1 += dt;
              }
          }
          if (r_antia) lastval = (byte)
            ( GetTexel( (t1>>8)&0x3, base, x1, xmask, z1, zmask )
             * (0xff-(t1&0xff)) >> 16 )
          + ( GetTexel( ((t1>>8)+1)&0x3, base, x1, xmask, z1, zmask )
             * (t1&0xff) >> 16 );
  }
  if ((r_antia) && (aac!=-1))
    vga->page_draw[offs] = lastval+((vga->page_draw[offs]-lastval)*aac>>8);
};

/*
 * the main drawing procedure
 */
void Draw_Voxel()
{
// declare variables
    long x, OldC, OldX, OldZ, NewC, RemC, NewX, NewZ, l;
    long NewY, OldY, RemY, RemX, RemZ;
    vector LeftRay, RightRay;
    long oY, oX, oZ, cY, cX, cZ, dcY, dcX, dcZ, rX, rZ, drX, drZ;
    bool SkipedVoxel;
    byte Startx, Stepx;

// get both rays at extrema
     LeftRay = normalize( Inverse * vector( (double)-160/HScale, 0, 1 ) ) * RenderZone;
     RightRay = normalize( Inverse * vector( (double)159/HScale, 0, 1 ) ) * RenderZone;
// get first ray's coordinates
     rX = (long)(LeftRay[0] * 65536 / ( RenderSteps * HeightMapScale ));
     rZ = (long)(LeftRay[2] * 65536 / ( RenderSteps * HeightMapScale ));
// get origin of rays in 16.16 fixed point
     oX = (long)(CamPosVoxSpace[0] * 65536 / HeightMapScale);
     oZ = (long)(CamPosVoxSpace[2] * 65536 / HeightMapScale);
// 8.8 fixed point for oY
     oY = (long)(CamPosVoxSpace[1] * 256);
// calculate deltas in 16.16 fixed point
     drX = (long)((RightRay[0] - LeftRay[0])*65536 / (HeightMapScale*320*RenderSteps));
     drZ = (long)((RightRay[2] - LeftRay[2])*65536 / (HeightMapScale*320*RenderSteps));
// handle interlace
     if (r_int) {
        Startx = Interlace;
        Stepx = 2;
        if (Interlace) {
           rX += drX;
           rZ += drZ;
        }
        drX <<= 1;
        drZ <<= 1;
     }
     else {
        Startx = 0;
        Stepx = 1;
     }
// loop for all colunns on screen
     for ( x=Startx; x<320; x+=Stepx )
     {
     // delta to scan along current ray
         dcX = rX;
         dcZ = rZ;
         dcY = (long)(-Camera[2][1] * RenderZone * 256 ) / RenderSteps;
     // get origin of current ray
         cX = oX;
         cZ = oZ;
         cY = oY;
     // set old voxel's info to position 0
         l = 0;
         OldX = cX;
         OldZ = cZ;
         OldY = ((Get_Value( HeightMap, cX, cZ )-cY) * DistInfo[l].Persp + (VCenter<<12));
         OldC = Get_Value( LightMap, cX, cZ );
     // set visible voxel on screen to lowest possible
         RemY = 200<<12;
         RemX = cX;
         RemZ = cZ;
         RemC = OldC;
         SkipedVoxel = true;
     // loop along the ray to cover zone
         while ((RemY>0) && (l<RenderSteps))
         {
         // get coordinate of next voxel
               cX += dcX*DistInfo[l].StepSize>>8;
               cY += dcY*DistInfo[l].StepSize>>8;
               cZ += dcZ*DistInfo[l].StepSize>>8;
         // project height of current voxel onto the screen
               NewY = (Get_Value( HeightMap, cX, cZ )-cY) * DistInfo[l].Persp + (VCenter<<12);
         // get current position in height map
               NewX = cX;
               NewZ = cZ;
         // get color with optional fog
               if (r_fog)
               {
                  if (r_light) NewC = (Get_Value( LightMap, cX, cZ )*DistInfo[l].FogVal)>>15;
                  else NewC = DistInfo[l].FogVal;
               } else NewC = Get_Value( LightMap, cX, cZ );
         // if this voxel is visible, then draw it
               if (NewY<RemY)
               {
               // clip if necessary, and readjust coordinates
                  if (NewY<0)
                  {
                     long coef = ((-OldY)<<8)/(NewY-OldY);
                     NewC = OldC + (long)((NewC-OldC) * coef >> 8 );
                     NewX = OldX + (long)((NewX-OldX) * coef >> 8 );
                     NewZ = OldZ + (long)((NewZ-OldZ) * coef >> 8 );
                     NewY = 0;
                  };
               // did we skip any voxels?
                  if (SkipedVoxel)
                  {
                     SkipedVoxel = false;
                  // if so, we must clip coordinates
                     long coef = ((RemY-OldY)<<8)/(NewY-OldY);
                     RemC = OldC + (long)((NewC-OldC) * coef >> 8 );
                     RemX = OldX + (long)((NewX-OldX) * coef >> 8 );
                     RemZ = OldZ + (long)((NewZ-OldZ) * coef >> 8 );
                     vLine( x, NewY, NewC, NewX, NewZ, RemY, RemC, RemX, RemZ, (0xff-(RemY>>4)&0xff) );
                  }
                  else  // draw line
                    vLine( x, NewY, NewC, NewX, NewZ, RemY, RemC, RemX, RemZ, -1 );
               // reset minimum height
                  RemY = NewY;
                  RemC = NewC;
                  RemX = NewX;
                  RemZ = NewZ;
              }
              else SkipedVoxel = true;
         // set new current position
               OldY = NewY;
               OldC = NewC;
               OldX = NewX;
               OldZ = NewZ;
         // increment distance from viewer
               l++;
         }
     // update cBuffer
         cBuffer[x] = RemY;
     // interpolate to get next ray
         rX += drX;
         rZ += drZ;
     }
};

/*
 * draws a textured ceiling "a la Doom"
 */
void Draw_Sky()
{
     word i, j;
     word Value, PixelNotDrawn;
     vector LeftRay, RightRay;
     double tr, tl, oX, oY, oZ;
     long xl, xr, zl, zr, dx, dz, offs, yv;
     byte pixel;
     byte startx, stepx;
     if (r_int)
     {
        stepx = 2;
        startx = Interlace;
        offs = Interlace;
     }
     else {
        stepx = 1;
        startx = 0;
        offs = 0;
     }
     oX = CamPosVoxSpace[0];
     oY = CamPosVoxSpace[1];
     oZ = CamPosVoxSpace[2];
     j = 0;
     PixelNotDrawn = 0;
// draw all pixels not drawn covered by the voxel
     while ((j<200) && (PixelNotDrawn<320))
     {
         // get intersection of ray with the plane
         tl = -1;
         tr = -1;
         LeftRay = Inverse * SkyRays[j][0];
         RightRay = Inverse * SkyRays[j][1];
         if (fabs(LeftRay[1])>0.1) tl = (SkyHeight-oY) / LeftRay[1];
         if (fabs(RightRay[1])>0.1) tr = (SkyHeight-oY) / RightRay[1];
         PixelNotDrawn = 0;
      // draw the horizontal line if it's not too far
         if ((tr>0) && (tr<1023))
         {
         // get texture coordinates
            xl = (long)(SkyScale * (oX + tl * LeftRay[0]));
            zl = (long)(SkyScale * (oZ + tl * LeftRay[2]));
            xr = (long)(SkyScale * (oX + tr * RightRay[0]));
            zr = (long)(SkyScale * (oZ + tr * RightRay[2]));
         // prepare interpolation
            dx = (long)(xr - xl)/320;
            dz = (long)(zr - zl)/320;
            if (r_int)
            {
               dx <<= 1;
               dz <<= 1;
            };
         // move the sky over time
            xl += timer->getCount()<<12;
         // draw the sky at full intensity color
            if ((tr<768) || (!r_fog))
            {
            // draw horizontal span
               for ( i=startx; i<320; i+=stepx )
               {
                  yv = (cBuffer[i]>>12);
             // check cBuffer, do not overdraw
                  if (j<yv)
                  {
                     vga->page_draw[offs] = (GetValue( SkyText, zl, xl )>>7);
                  } else PixelNotDrawn+=stepx;
                  if ((j==yv) && (r_antia)) {
                     pixel = (GetValue( SkyText, zl, xl )>>7);
                     vga->page_draw[offs] = pixel+((vga->page_draw[offs]-pixel)*(0xff-(cBuffer[i]>>8)&0xff)>>8);
                  }
                  offs+=stepx;
             // interpolate
                  xl += dx;
                  zl += dz;
              };
            }
         // draw faded sky in the distance
            else {
           // get fade value
              Value = (word)(1023-tr)<<8;
           // draw h-line
              for ( i=startx; i<320; i+=stepx )
              {
                yv = (cBuffer[i]>>12);
             // check cBuffer, prevent overdraw
                if (j<yv)
                {
                   vga->page_draw[offs] = PropTab[Value+((GetValue( SkyText, zl, xl )>>7)&0xff)];
                } else PixelNotDrawn+=stepx;
                if ((j==yv) && (r_antia)) {
                   pixel = PropTab[Value+((GetValue( SkyText, zl, xl )>>7)&0xff)];;
                   vga->page_draw[offs] = pixel+((vga->page_draw[offs]-pixel)*(0xff-(cBuffer[i]>>8)&0xff)>>8);
                }
                offs+=stepx;
             // interpolate
                xl += dx;
                zl += dz;
              };
            }
         }
      // just draw a unicolour horizontal line
         else {
                for ( i=startx; i<320; i+=stepx )
                {
                    yv = (cBuffer[i]>>12);
                    if (j<yv)
                    {
                      vga->page_draw[offs] = 0;
                    } else PixelNotDrawn+=stepx;
                    if ((j==yv) && (r_antia))
                        vga->page_draw[offs] = (vga->page_draw[offs])*(0xff-(cBuffer[i]>>8)&0xff)>>8;
                    offs+=stepx;
                };
             };
      // do next line only if some pixels need drawing
         j++;
     };
};

/*
 * convert a linearily accessible map to a blocked map
 */
void Make_Blocks( byte *where )
{
     byte *tmp = (byte *)malloc( 1048576 );
     int i, j, x, y, offs;
     offs = 0;
     for (j=0; j<(1024/BlockSizeY); j++)
       for (i=0; i<(1024/BlockSizeX); i++)
         for (y=0; y<BlockSizeY; y++)
           for (x=0; x<BlockSizeX; x++)
           {
               tmp[offs] = where[((j*BlockSizeY+y)<<10)+(i*BlockSizeX+x)];
               offs++;
           }
     memcpy( where, tmp, 1048576 );
     free( tmp );
}


/*
 * get the position of the camera in voxel space
 * this procedure is a bit of a hack! there must be a better way
 */
void Find_Inverse_Transformation()
{
   CamPosVoxSpace = Camera[3];
   Inverse = Camera;
   Inverse = transpose( Inverse );
   Inverse[3] = vector( 0, 0, 0 );
   Inverse[0][3] = 0;   Inverse[1][3] = 0;
   Inverse[2][3] = 0;   Inverse[3][3] = 1;
   CamPosVoxSpace = Inverse * -CamPosVoxSpace;
};

/*
 * allocate memory and load a raw file into it
 */
void load_raw( char *name, byte **buf, long size )
{
   if (!packfile->isOk()) return;
   *buf = (byte *)malloc( size );
   packfile->seek( name );
   packfile->read( *buf, size );
}

/*
 * main setup procedure
 */
void Init_Voxel()
{
     int i, j, k;
     cout << "-{ terraVox: an experimental voxel landscape engine }-\n\n";
     cout << "Please wait...\n";
// allocate memory
     MixTab     = (byte *)malloc( 65536 );
     PropTab    = (byte *)malloc( 65536 );
     LightMap   = (byte *)malloc( 1048576 );
     RefMap     = (byte *)malloc( 1048576 );
     cBuffer    = (long *)malloc( 1280 );
     DistInfo   = (TDistInfo *)malloc( RenderSteps*sizeof( TDistInfo ) );
// load height map and color map
     load_raw( "hmap.raw", &HeightMap, 1048576 );
// calculate reference map
     for (i=0; i<1048576; i++) {
          k = HeightMap[i] + (rand() & 128) - 64;
          if (k<0) k = 0;
          if (k>255) k = 255;
          LightMap[i] = k;
     }
     for (j=0; j<1024; j++)
         for (i=0; i<1024; i++)
         {
             RefMap[(j<<10)+i] = ( LightMap[(((j+1)<<10)&0xffc00)+((i+1)&0x003ff)]
                                 + LightMap[(((j+1)<<10)&0xffc00)+((i-1)&0x003ff)]
                                 + LightMap[(((j-1)<<10)&0xffc00)+((i+1)&0x003ff)]
                                 + LightMap[(((j-1)<<10)&0xffc00)+((i-1)&0x003ff)] ) >> 8;
         }
// compute lightmap
     for (j=0; j<1024; j++)
         for (i=0; i<1024; i++)
         {
             k = 140+((HeightMap[(j<<10)+((i+1)&0x3ff)]-HeightMap[(j<<10)+i])<<3);
             if (k<0) k = 0;
             if (k>255) k = 255;
             LightMap[(j<<10)+i] = (byte)k;
         }
// make maps into n x n blocks
     Make_Blocks( HeightMap );
     Make_Blocks( LightMap );
     Make_Blocks( RefMap );
// load sky texture
     load_raw( "sky.raw", &SkyText, 65536 );
// load landscape texture
     load_raw( "text1.raw", &Texture[0], 65536 );
     load_raw( "text2.raw", &Texture[1], 65536 );
     load_raw( "text3.raw", &Texture[2], 65536 );
     load_raw( "text4.raw", &Texture[3], 65536 );
// load some fonts and digits
     load_raw( "info.raw", &Info, 17640 );
     load_raw( "digits.raw", &Digits, 640 );
// prepare bilinear filtering
     Prepare_Bilinear();
// misc 3d setup
     Camera = translate( 0, 100, 128 * HeightMapScale );
     TrX = 0; TrY = 0; TrZ = 0; AngX = 0; AngY = 0; AngZ = 0;
// precalculate projection rays to map the sky quicker
     for (j=-100; j<100; j++)
     {
         SkyRays[j+100][0] = normalize( vector( (double)-160 / HScale, (double)j / VScale, 1 ) ) * 16;
         SkyRays[j+100][1] = normalize( vector( (double)159 / HScale,  (double)j / VScale, 1 ) ) * 16;
     };
// prepare fog interpolation
     long FogVal = NearFogValue;
     long FogInt = (FarFogValue-NearFogValue)/(RenderSteps-NearFogStep);
// prepare step size interpolation
     long StepSize = NearStepSize;
     long StepInt = (FarStepSize-NearStepSize)/RenderSteps;
     long RefStep = (RenderZone<<8)/RenderSteps;
     long Dist = 0;
     for (i=0; i<RenderSteps; i++)
     {
      // get next voxel
         Dist += (RefStep*StepSize)>>8;
         DistInfo[i].StepSize = StepSize;
      // perspective projection
         DistInfo[i].Persp = (long)( 4096 * VScale / Dist ); // (i*RenderZone/RenderSteps)
      // fog value
         DistInfo[i].FogVal = FogVal;
         if (i>=NearFogStep) FogVal += FogInt;
      // step shift
         StepSize += StepInt;
     };
};

/*
 * deallocate all memory
 */
void Done_Voxel()
{
     free( HeightMap  );
     free( LightMap   );
     free( RefMap     );
     free( MixTab     );
     free( PropTab    );
     free( SkyText    );
     free( Texture[0] );
     free( Texture[1] );
     free( Texture[2] );
     free( Texture[3] );
     free( Digits     );
     free( Info       );
     free( cBuffer    );
     free( DistInfo   );
}

void Set_Option( bool *opt, byte pos )
{
     *opt = !(*opt);
     optNum = pos * 14;
     optCount = timer->getCount() + 210;
}

/*
 *  reads the keyboard input
 */
void Get_User_Input()
{
// camera movement
  while (timer->getCount()-lastUpdt>2)
  {
     lastUpdt += 2;
     if (keyboard->isKeyPressed(79)) TrX -= 0.04;
     if (keyboard->isKeyPressed(81)) TrX += 0.04;
     if (!keyboard->isKeyPressed(81) && !keyboard->isKeyPressed(79)) TrX /= 1.05;
     if (keyboard->isKeyPressed(16)) TrZ += 0.04;
     if (keyboard->isKeyPressed(30)) TrZ -= 0.04;
     if (keyboard->isKeyPressed(75)) Camera = Camera * rotY( 0.005 );
     if (keyboard->isKeyPressed(77)) Camera = Camera * rotY( -0.005 );
     if (keyboard->isKeyPressed(17)) TrY += 0.004;
     if (keyboard->isKeyPressed(31)) TrY -= 0.004;
     if (!keyboard->isKeyPressed(17) && !keyboard->isKeyPressed(31)) TrY /= 1.05;
     if (keyboard->isKeyPressed(57)) {
        TrX /= 1.1; TrY /= 1.1; TrZ /= 1.1;
     }
     Camera = Camera * translate( TrX, TrY, TrZ );
  }
// options
     if (keyboard->getKey(59)) Set_Option( &r_light, 0 );
     if (keyboard->getKey(60)) Set_Option( &r_fog, 1 );
     if (keyboard->getKey(61)) Set_Option( &r_mip, 2 );
     if (keyboard->getKey(62)) Set_Option( &r_bili, 3 );
     if (keyboard->getKey(63)) Set_Option( &r_antia, 4 );
     if (keyboard->getKey(64)) Set_Option( &r_int, 5 );
}

void Put_Digit( int x, int y, byte n )
{
  int i, j;
  for (j=0; j<8; j++)
  {
      for (i=0; i<8; i++)
      {
          if (Digits[(n<<6)+(j<<3)+i]!=0) vga->page_draw[(y+j)*320+i+x] = 255;
      }
  }
}

/*
 * print text on the screen to try and tell the user
 * what options he's trying to set
 */
void Put_Info()
{
// show option information
   if (timer->getCount()<optCount)
   {
       int i, j;
       for (j=0; j<14; j++)
       {
           for (i=0; i<180; i++)
           {
               if (Info[((optNum+j)*180)+i]!=0) vga->page_draw[(186+j)*320+i] = Info[((optNum+j)*180)+i];
           }
        }
   }
// calculate FPS
   if (lastTime+70<timer->getCount())
   {
      curFPS = (FrameCount-lastCount) * 210 / (timer->getCount()-lastTime);
      lastTime = timer->getCount();
      lastCount = FrameCount;
   }
// show FPS
   Put_Digit( 16, 0, curFPS % 10 );
   Put_Digit( 8, 0, (curFPS/10) % 10 );
   if ((curFPS/100) % 10!=0) Put_Digit( 0, 0, (curFPS/100) % 10 );
}

/*
 * the main procedure
 */
int main()
{
// init everything
    packfile = new PACKFILE( "data.pak" );
    Init_Voxel();
    if (!packfile->isOk()) {
         cout << "\nError with pack file.\n";
         return 1;
    }
    timer = new TIMER( 210 );
    keyboard = new KEYBOARD;
    vga = new VGA;
// run main loop
    while (!keyboard->isKeyPressed(1))
    {
           Get_User_Input();
           Camera = Camera * translate( TrX, TrY, TrZ );
           Find_Inverse_Transformation();
           Reset_cBuffer();
           if (CamPosVoxSpace[1]<SkyHeight) Draw_Voxel();
           Draw_Sky();
           Put_Info(); // aaarrggg!! OVERDRAW :)
           vga->Update();
           FrameCount++;
           Interlace = 1-Interlace;
    }
// un-init
  delete keyboard;
  delete vga;
  delete timer;
  delete packfile;
  Done_Voxel();
// all done
  cout << "-{ terraVox: an experimental voxel landscape engine }-\n\n";
  cout << "   * Code:      Alex Champandard\n";
  cout << "   * Contact:   alexjc@altavista.com\n\n";
  cout << "Check http://atlas.cs.york.ac.uk/~ajc116/terraVox for updates.\n";
  return 0;
};
